r"""
Most integration problems when dealing with polynomial chaos expansion comes
with a weight function :math:`p(x)` which happens to be the probability density
function. Gaussian quadrature creates weights and abscissas that are tailored
to be optimal with the inclusion of a weight function. It is therefore not one
method, but a collection of methods, each tailored to different probability
density functions.

In ``chaospy`` Gaussian quadrature is a functionality attached to each
probability distribution. This means that instead of explicitly supporting
a list of quadrature rules, all rules are supported through the capability of
the distribution implementation. For common distribution, this means that the
quadrature rules are calculated analytically using Stieltjes method on known
three terms recursion coefficients, and using those to create quadrature node
using the Golub-Welsch algorithm.

For example for the tailored quadrature rules defined above:

* Gauss-Hermit quadrature is tailored to the normal (Gaussian) distribution::

    >>> X, W = chaospy.generate_quadrature(4, chaospy.Normal(0, 1), rule="G")
    >>> print(numpy.around(X, 3), numpy.around(W, 3))
    [[-2.857 -1.356 -0.     1.356  2.857]] [0.011 0.222 0.533 0.222 0.011]

* Gauss-Legendre quadrature is tailored to the Uniform distributions::

    >>> X, W = chaospy.generate_quadrature(4, chaospy.Uniform(0, 1), rule="G")
    >>> print(numpy.around(X, 3), numpy.around(W, 3))
    [[0.047 0.231 0.5   0.769 0.953]] [0.118 0.239 0.284 0.239 0.118]

* Gauss-Jacobi quadrature is tailored to the Beta distribution::

    >>> X, W = chaospy.generate_quadrature(4, chaospy.Beta(2, 4), rule="G")
    >>> print(numpy.around(X, 3), numpy.around(W, 3))
    [[0.067 0.212 0.41  0.627 0.827]] [0.118 0.367 0.36  0.139 0.015]

* Gauss-Laguerre quadrature is tailored to the Exponential distribution::

    >>> X, W = chaospy.generate_quadrature(4, chaospy.Exponential(), rule="G")
    >>> print(numpy.around(X, 3), numpy.around(W, 3))
    [[ 0.264  1.413  3.596  7.086 12.641]] [0.522 0.399 0.076 0.004 0.   ]

* Generalized Gauss-Laguerre quadrature is tailored to the Gamma distribution::

    >>> X, W = chaospy.generate_quadrature(4, chaospy.Gamma(2, 4), rule="G")
    >>> print(numpy.around(X, 3), numpy.around(W, 3))
    [[ 2.468  8.452 18.443 33.596 57.04 ]] [0.348 0.502 0.141 0.009 0.   ]

For uncommon distributions an analytical Stieltjes method can not be performed
as the distribution does not provide three terms recursion coefficients. In
this scenario, the discretized counterpart is used instead as an approximation.
For example, to mention a few:

* The triangle distribution::

    >>> X, W = chaospy.generate_quadrature(4, chaospy.Triangle(), rule="G")
    >>> print(numpy.around(X, 3), numpy.around(W, 3))
    [[-0.821 -0.45  -0.     0.45   0.821]] [0.052 0.239 0.418 0.239 0.052]

* The Laplace distribution::

    >>> X, W = chaospy.generate_quadrature(4, chaospy.Laplace(), rule="G")
    >>> print(numpy.around(X, 3), numpy.around(W, 3))
    [[-8.091 -2.855 -0.     2.855  8.091]] [0.001 0.114 0.771 0.114 0.001]

* The Weibull distribution::

    >>> X, W = chaospy.generate_quadrature(4, chaospy.Weibull(), rule="G")
    >>> print(numpy.around(X, 3), numpy.around(W, 3))
    [[ 0.264  1.413  3.596  7.086 12.64 ]] [0.522 0.399 0.076 0.004 0.   ]

* The Rayleigh distribution::

    >>> X, W = chaospy.generate_quadrature(4, chaospy.Rayleigh(), rule="G")
    >>> print(numpy.around(X, 3), numpy.around(W, 3))
    [[0.308 0.938 1.779 2.79  4.032]] [0.144 0.453 0.339 0.063 0.002]

As a small side note, it is worth noting that since the weight function is
assumed to be a probability density function, we here only focuses on
probabilistic Gaussian quadrature. This means that we assume the constraint:

.. math::

    \int p(x) dx = 1

There is also another version, often named as the physicist version of Gaussian
quadrature. They are for the most part the same, but the latter have other
weighting constraints. To jump between the two variants, most often one has to
multiply the weights with an appropriate constant to achieve the correct
values.
"""
import numpy
import scipy.linalg

import chaospy.quad


def quad_golub_welsch(order, dist, accuracy=100, **kws):
    """
    Golub-Welsch algorithm for creating quadrature nodes and weights.

    Args:
        order (int):
            Quadrature order
        dist (Dist):
            Distribution nodes and weights are found for with `dim=len(dist)`
        accuracy (int):
            Accuracy used in discretized Stieltjes procedure. Will
            be increased by one for each iteration.

    Returns:
        (numpy.ndarray, numpy.ndarray):
            Optimal collocation nodes with `x.shape=(dim, order+1)` and weights
            with `w.shape=(order+1,)`.

    Examples:
        >>> Z = chaospy.Normal()
        >>> x, w = chaospy.quad_golub_welsch(3, Z)
        >>> print(numpy.around(x, 4))
        [[-2.3344 -0.742   0.742   2.3344]]
        >>> print(numpy.around(w, 4))
        [0.0459 0.4541 0.4541 0.0459]
        >>> Z = chaospy.J(chaospy.Uniform(), chaospy.Uniform())
        >>> x, w = chaospy.quad_golub_welsch(1, Z)
        >>> print(numpy.around(x, 4))
        [[0.2113 0.2113 0.7887 0.7887]
         [0.2113 0.7887 0.2113 0.7887]]
        >>> print(numpy.around(w, 4))
        [0.25 0.25 0.25 0.25]
    """
    order = numpy.array(order)*numpy.ones(len(dist), dtype=int)+1
    _, _, coeff1, coeff2 = chaospy.quad.generate_stieltjes(
        dist, numpy.max(order), accuracy=accuracy, retall=True, **kws)

    dimensions = len(dist)
    abscisas, weights = _golbub_welsch(order, coeff1, coeff2)

    if dimensions == 1:
        abscisa = numpy.reshape(abscisas, (1, order[0]))
        weight = numpy.reshape(weights, (order[0],))
    else:
        abscisa = chaospy.quad.combine(abscisas).T
        weight = numpy.prod(chaospy.quad.combine(weights), -1)

    assert len(abscisa) == dimensions
    assert len(weight) == len(abscisa.T)
    return abscisa, weight


def _golbub_welsch(orders, coeff1, coeff2):
    """Recurrence coefficients to abscisas and weights."""
    abscisas, weights = [], []

    for dim, order in enumerate(orders):
        if order:
            bands = numpy.zeros((2, order))
            bands[0] = coeff1[dim, :order]
            bands[1, :-1] = numpy.sqrt(coeff2[dim, 1:order])
            vals, vecs = scipy.linalg.eig_banded(bands, lower=True)

            abscisa, weight = vals.real, vecs[0, :]**2
            indices = numpy.argsort(abscisa)
            abscisa, weight = abscisa[indices], weight[indices]

        else:
            abscisa, weight = numpy.array([coeff1[dim, 0]]), numpy.array([1.])

        abscisas.append(abscisa)
        weights.append(weight)
    return abscisas, weights



import os
import shutil
import subprocess

from cubature import cubature
from test_cubature import main

__version__ = '0.12.0'

def run_test(ndim=3, tol=1.e-5, functions=[0,1,2,3,4,5,6,7],
             maxEval=1000000, fdim=5):
    '''Default test for the Cubature package.

    Parameters
    ----------

    ndim : integer
        Number of dimensions to integrate over.
        (default = 3)

    tol : float
        Error tolerance for the test
        (default = 1.e-5)

    functions : list or tuple
        (default = [0,1,2,3,4,5,6,7])
        The different test functions are:
        0 - a product of cosine functions
        1 - a Gaussian integral of exp(-x^2), remapped to [0,infinity) limits
        2 - volume of a hypersphere (integrating a discontinuous function!)
        3 - a simple polynomial (product of coordinates)
        4 - a Gaussian centered in the middle of the integration volume
        5 - a sum of two Gaussians
        6 - an example function by Tsuda, a product of terms with near poles
        7 - a test integrand by Morokoff and Caflisch, a simple product of
           dim-th roots of the coordinates (weakly singular at the boundary)
    maxEval : integer
        Maximum number of function evaluations.
        (default = 1000000)
    fdim : integer
        Size for the vector-valued function.
        (default = 5)
    '''
    cwd = os.getcwd()
    main(ndim, tol, functions, maxEval, fdim)



# pylint: disable=wildcard-import
"""
To create quadrature abscissas and weights, use the
:func:`~chaospy.quad.generate_quadrature` function. Which type of quadrature to
use is defined by the flag ``rule``. This argument can either be the full name,
or a single letter representing the rule. These are as follows.

``Gaussian``, ``G``
    Optimal Gaussian quadrature from Golub-Welsch. A slow method for higher
    order.
``Legendre``, ``E``
    Gauss-Legendre quadrature
``Clenshaw``, ``C``
    Clenshaw-Curtis quadrature. Exponential growth rule is used when sparse is
    True to make the rule nested.
``Leja``, J``
    Leja quadrature. Linear growth rule is nested.
``Genz``, ``Z``
    Hermite Genz-Keizter 16 rule. Nested. Valid to order 8.
``Patterson``, ``P``
    Gauss-Patterson quadrature rule. Nested. Valid to order 8.
``Fejer``, ``F``
    Fejér quadrature. Same as Clenshaw-Curtis, but without the endpoints.
"""
from .frontend import *



import logging
import inspect

import networkx
import numpy

from ... import quad

from .common import DependencyError
from .parameters import load_parameters as load_inputs

from .density import evaluate_density
from .forward import evaluate_forward
from .inverse import evaluate_inverse
from .bound import evaluate_bound
from .moment import evaluate_moment
from .recurrence_coefficients import evaluate_recurrence_coefficients




def sorted_dependencies(dist, reverse=False):
    from .. import baseclass
    graph = networkx.DiGraph()
    graph.add_node(dist)
    dist_collection = [dist]
    while dist_collection:

        cur_dist = dist_collection.pop()

        values = (value for value in cur_dist.prm.values()
                  if isinstance(value, baseclass.Dist))
        for value in values:
            if value not in graph.node:
                graph.add_node(value)
                dist_collection.append(value)
            graph.add_edge(cur_dist, value)

    if not networkx.is_directed_acyclic_graph(graph):
        raise DependencyError("cycles in dependency structure.")

    sorting = list(networkx.topological_sort(graph))
    if reverse:
        sorting = reversed(sorting)

    for dist in sorting:
        yield dist


def get_dependencies(*distributions):
    from .. import baseclass
    distributions = [
        set(sorted_dependencies(dist)) for dist in distributions
        if isinstance(dist, baseclass.Dist)
    ]
    if len(distributions) <= 1:
        return set()
    intersections = set.intersection(*distributions)
    return intersections




def get_forward_cache(
        distribution,
        cache,
):
    from .. import baseclass
    if not isinstance(distribution, baseclass.Dist):
        return distribution
    if distribution in cache:
        return cache[distribution]
    if hasattr(distribution, "_fwd_cache"):
        return distribution._fwd_cache(cache)
    return distribution

def get_inverse_cache(
        distribution,
        cache,
):
    from .. import baseclass
    if not isinstance(distribution, baseclass.Dist):
        return distribution
    if distribution in cache:
        return cache[distribution]
    if hasattr(distribution, "_inv_cache"):
        return distribution._inv_cache(cache)
    return distribution



from distutils.core import setup
from distutils.extension import Extension
try:
    from Cython.Distutils import build_ext
    ext_modules = [Extension('_cubature', ['_cubature.pyx'])]
    cmdclass = {'build_ext': build_ext}
except ImportError:
    ext_modules = [Extension('_cubature', ['_cubature.c'])]
    cmdclass = {}
    pass
setup(
cmdclass = cmdclass,
ext_modules = ext_modules,
)



"""
Quadrature methods, or numerical integration, is broad class of algorithm for
performing integration of any function :math:`g` that are defined without
requiring an analytical definition. In the scope of ``chaospy`` we limit this
scope to focus on methods that can be reduced to the following approximation:

.. math::
    \int g(x) p(x) dx = \sum_{n=1}^N W_n g(X_n)

Here :math:`p(x)` is an weight function, which for our use would be an
probability distribution, and :math:`W_n` and :math:`X_n` are respectively
quadrature weights and abscissas used to define the approximation.

This simplest example of such an approximation is Monte Carlo integration. In
such a method, you only need to select :math:`W_n=1/N` and :math:`X_n` to be
independent identical distributed samples drawn from the distribution of
:math:`p(x)`.

However, except for very high dimensional problems, Monte Carlo is quite an
inefficient way to perform numerical integration, and there exist quite a few
methods that performs better in most low-dimensional settings.
"""
from .combine import combine
from .interface import generate_quadrature

from .stieltjes import generate_stieltjes
from .collection import *
from .sparse_grid import sparse_grid
from .generator import rule_generator



"""Multivariate Normal Distribution."""
import numpy
from scipy import special, misc

from .normal import normal

from ..baseclass import Dist


class MvNormal(Dist):
    """
    Multivariate Normal Distribution.

    Args:
        loc (float, Dist): Mean vector
        scale (float, Dist): Covariance matrix or variance vector if scale is a 1-d vector.

    Examples:
        >>> distribution = chaospy.MvNormal([1, 2], [[1, 0.6], [0.6, 1]])
        >>> print(distribution)
        MvNormal(loc=[1.0, 2.0], scale=[[1.0, 0.6], [0.6, 1.0]])
        >>> mesh = numpy.meshgrid(*[numpy.linspace(0, 1, 5)[1:-1]]*2)
        >>> print(numpy.around(distribution.inv(mesh), 4))
        [[[0.3255 1.     1.6745]
          [0.3255 1.     1.6745]
          [0.3255 1.     1.6745]]
        <BLANKLINE>
         [[1.0557 1.4604 1.8651]
          [1.5953 2.     2.4047]
          [2.1349 2.5396 2.9443]]]
        >>> print(numpy.around(distribution.fwd(distribution.inv(mesh)), 4))
        [[[0.25 0.5  0.75]
          [0.25 0.5  0.75]
          [0.25 0.5  0.75]]
        <BLANKLINE>
         [[0.25 0.25 0.25]
          [0.5  0.5  0.5 ]
          [0.75 0.75 0.75]]]
        >>> print(numpy.around(distribution.pdf(distribution.inv(mesh)), 4))
        [[0.0991 0.146  0.1452]
         [0.1634 0.1989 0.1634]
         [0.1452 0.146  0.0991]]
        >>> print(numpy.around(distribution.sample(4), 4))
        [[ 1.395  -0.2003  2.6476  0.9553]
         [ 3.1476  0.6411  1.5946  1.7647]]
        >>> print(numpy.around(distribution.mom((1, 2)), 4))
        7.4
        >>> print(numpy.around(distribution.ttr([[1, 1, 2], [1, 2, 2]]), 4))
        [[[1.   1.   1.  ]
          [1.36 1.36 2.72]]
        <BLANKLINE>
         [[2.   2.   2.  ]
          [0.64 1.28 1.28]]]
    """

    def __init__(self, loc=[0, 0], scale=[[1, .5], [.5, 1]]):
        loc = numpy.asfarray(loc)
        scale = numpy.asfarray(scale)
        assert len(loc) == len(scale)
        self._repr = {"loc": loc.tolist(), "scale": scale.tolist()}

        C = numpy.linalg.cholesky(scale)
        Ci = numpy.linalg.inv(C)
        Dist.__init__(self, C=C, Ci=Ci, loc=loc)

    def _cdf(self, x, C, Ci, loc):
        return special.ndtr(numpy.dot(Ci, (x.T-loc.T).T))

    def _ppf(self, q, C, Ci, loc):
        return (numpy.dot(C, special.ndtri(q)).T+loc.T).T

    def _pdf(self, x, C, Ci, loc):
        det = numpy.linalg.det(numpy.dot(C,C.T))
        x_ = numpy.dot(Ci.T, (x.T-loc.T).T)
        out = numpy.ones(x.shape)
        out[0] =  numpy.e**(-.5*numpy.sum(x_*x_, 0))/numpy.sqrt((2*numpy.pi)**len(Ci)*det)
        return out

    def _bnd(self, x, C, Ci, loc):
        scale = numpy.sqrt(numpy.diag(numpy.dot(C,C.T)))
        lo,up = numpy.zeros((2,)+x.shape)
        lo.T[:] = (-7.5*scale+loc)
        up.T[:] = (7.5*scale+loc)
        return lo, up

    def _mom(self, k, C, Ci, loc):
        scale = numpy.dot(C, C.T)

        def mom(k):

            zeros = (numpy.sum(k,0)%2==1)+numpy.any(numpy.array(k)<0, 0)
            if numpy.all(zeros, 0):
                return 0.

            dim, K = k.shape
            ra = numpy.arange(dim).repeat(K).reshape(dim,K)

            i = numpy.argmax(k != 0, 0)

            out = numpy.zeros(k.shape[1:])
            out[:] = numpy.where(numpy.choose(i,k),
                    (numpy.choose(i,k)-1)*scale[i,i]*mom(k-2*(ra==i)), 1)
            for x in range(1, dim):
                out += \
                (numpy.choose(i,k)!=0)*(x>i)*k[x]*scale[i,x]*mom(k-(ra==i)-(ra==x))

            return out

        dim = len(loc)
        K = numpy.mgrid[[slice(0,_+1,1) for _ in k]]
        K = K.reshape(dim, -1)
        M = mom(K)
        out = 0.

        for i in range(len(M)):
            coef = numpy.prod(misc.comb(k.T, K[:,i]).T, 0)
            diff = k.T - K[:,i]
            pos = diff >= 0
            diff = diff*pos
            pos = numpy.all(pos)
            loc_ = numpy.prod(loc**diff)
            out += pos*coef*loc_*M[i]

        return out

    def __len__(self):
        return len(self.prm["C"])



"""Raised cosine distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add
from .deprecate import deprecation_warning


class raised_cosine(Dist):
    """Raised cosine distribution."""

    def __init__(self):
        Dist.__init__(self)

    def _pdf(self, x):
        return .5 + .5*numpy.cos(numpy.pi*x)

    def _cdf(self, x):
        return .5 + .5*x + numpy.sin(numpy.pi*x)/(2*numpy.pi)

    def _bnd(self, x):
        return -1,1

    def _mom(self, k):
        output = numpy.array([special.hyp1f2(k_+.5, .5, k_+1.5, -numpy.pi**2/4)[0]
                              for k_ in k.flatten()]).reshape(k.shape)
        output = 1/(1.+k) + 1/(1.+2*k)*output
        output = numpy.where(k % 2, output, 0)
        return output


class RaisedCosine(Add):
    """
    Raised cosine distribution

    Args:
        loc (float, Dist) : Location parameter
        scale (float, Dist) : Scale parameter

    Examples:
        >>> distribution = chaospy.RaisedCosine(2, 2)
        >>> print(distribution)
        RaisedCosine(loc=2, scale=2)
        >>> q = numpy.linspace(0, 1, 7)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [1.2535 1.6586 2.     2.3414 2.7465]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.1667 0.3333 0.5    0.6667 0.8333]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.3469 0.4649 0.5    0.4649 0.3469]
        >>> print(numpy.around(distribution.sample(4), 4))
        [2.3134 1.0903 3.1938 1.9644]
        >>> print(numpy.around(distribution.mom(1), 4))
        2.5947
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[2.     2.     2.    ]
         [0.5228 0.7351 0.8339]]
    """

    def __init__(self, loc=0, scale=1):
        self._repr = {"loc": loc, "scale": scale}
        Add.__init__(self, left=raised_cosine()*scale, right=loc)


Raised_cosine = deprecation_warning(RaisedCosine, "Raised_cosine")



"""Triangle probability distribution."""
import numpy
from scipy import special, misc

from ..baseclass import Dist
from ..operators.addition import Add
from .beta import beta_


def tri_ttr(k, a):
    """
    Custom TTR function.

    Triangle distribution does not have an analytical TTR function, but because
    of its non-smooth nature, a blind integration scheme will converge very
    slowly. However, by splitting the integration into two divided at the
    discontinuity in the derivative, TTR can be made operative.
    """
    from ...quad import quad_clenshaw_curtis
    q1, w1 = quad_clenshaw_curtis(int(10**3*a), 0, a)
    q2, w2 = quad_clenshaw_curtis(int(10**3*(1-a)), a, 1)
    q = numpy.concatenate([q1,q2], 1)
    w = numpy.concatenate([w1,w2])
    w = w*numpy.where(q<a, 2*q/a, 2*(1-q)/(1-a))

    from chaospy.poly import variable
    x = variable()

    orth = [x*0, x**0]
    inner = numpy.sum(q*w, -1)
    norms = [1., 1.]
    A,B = [],[]

    for n in range(k):
        A.append(inner/norms[-1])
        B.append(norms[-1]/norms[-2])
        orth.append((x-A[-1])*orth[-1]-orth[-2]*B[-1])

        y = orth[-1](*q)**2*w
        inner = numpy.sum(q*y, -1)
        norms.append(numpy.sum(y, -1))

    A, B = numpy.array(A).T[0], numpy.array(B).T
    return A[-1], B[-1]


class triangle(Dist):
    """Triangle probability distribution."""

    def __init__(self, a=.5):
        assert numpy.all(a>=0) and numpy.all(a<=1)
        Dist.__init__(self, a=a)

    def _pdf(self, D, a):
        return numpy.where(D<a, 2*D/a, 2*(1-D)/(1-a))

    def _cdf(self, D, a):
        return numpy.where(D<a, D**2/(a + (a==0)),
                (2*D-D*D-a)/(1-a+(a==1)))

    def _ppf(self, q, a):
        return numpy.where(q<a, numpy.sqrt(q*a), 1-numpy.sqrt(1-a-q*(1-a)))

    def _mom(self, k, a):
        a_ = a*(a!=1)
        out = 2*(1.-a_**(k+1))/((k+1)*(k+2)*(1-a_))
        return numpy.where(a==1, 2./(k+2), out)

    def _bnd(self, x, a):
        return 0., 1.

    def _ttr(self, k, a):
        if a == 0:
            return beta_()._ttr(k, 1, 2)
        if a == 1:
            return beta_()._ttr(k, 2, 1)
        return tri_ttr(k.item()+1, a)


class Triangle(Add):
    """
    Triangle Distribution.

    Must have lower <= midpoint <= upper.

    Args:
        lower (float, Dist) : Lower bound
        midpoint (float, Dist) : Location of the top
        upper (float, Dist) : Upper bound

    Examples:
        >>> distribution = chaospy.Triangle(2, 3, 4)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [2.6325 2.8944 3.1056 3.3675]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.6325 0.8944 0.8944 0.6325]
        >>> print(numpy.around(distribution.sample(4), 4))
        [3.1676 2.4796 3.6847 2.982 ]
        >>> print(numpy.around(distribution.mom(1), 4))
        3.0
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[3.     3.     3.    ]
         [0.1667 0.2333 0.2327]]
    """

    def __init__(self, lower=-1, midpoint=0, upper=1):
        self._repr = {"lower": lower, "midpoint": midpoint, "upper": upper}
        midpoint = (midpoint-lower)*1./(upper-lower)
        Add.__init__(self, left=triangle(midpoint)*(upper-lower), right=lower)



"""
The Gauss-Legendre quadrature rule is properly supported by in
:ref:`gaussian_quadrature`. However, as Gauss-Legendre is a special case where
the weight function is constant, it can in principle be used to integrate any
weighting function. In other words, this is the same Gauss-Legendre integration
rule, but only in the context of uniform distribution as weight function.
Normalization of the weights will be used to achieve the general integration
form.

It is also worth noting that this specific implementation of Gauss-Legendre is
faster to compute than the general version in :ref:`gaussian_quadrature`.

Example usage
-------------

The first few orders::

    >>> distribution = chaospy.Uniform(0, 1)
    >>> for order in [0, 1, 2, 3]:
    ...     abscissas, weights = chaospy.generate_quadrature(
    ...         order, distribution, rule="E")
    ...     print(order, numpy.around(abscissas, 3), numpy.around(weights, 3))
    0 [[0.5]] [1.]
    1 [[0.211 0.789]] [0.5 0.5]
    2 [[0.113 0.5   0.887]] [0.278 0.444 0.278]
    3 [[0.069 0.33  0.67  0.931]] [0.174 0.326 0.326 0.174]

Using an alternative distribution::

    >>> distribution = chaospy.Beta(2, 4)
    >>> for order in [0, 1, 2, 3]:
    ...     abscissas, weights = chaospy.generate_quadrature(
    ...         order, distribution, rule="E")
    ...     print(order, numpy.around(abscissas, 3), numpy.around(weights, 3))
    0 [[0.5]] [1.]
    1 [[0.211 0.789]] [0.933 0.067]
    2 [[0.113 0.5   0.887]] [0.437 0.556 0.007]
    3 [[0.069 0.33  0.67  0.931]] [0.195 0.647 0.157 0.001]

The abscissas stays the same, but the weights are re-adjusted for the new
weight function.
"""
import numpy

import chaospy.quad


def quad_gauss_legendre(order, lower=0, upper=1, composite=None):
    """
    Generate the quadrature nodes and weights in Gauss-Legendre quadrature.

    Example:
        >>> abscissas, weights = quad_gauss_legendre(3)
        >>> print(numpy.around(abscissas, 4))
        [[0.0694 0.33   0.67   0.9306]]
        >>> print(numpy.around(weights, 4))
        [0.1739 0.3261 0.3261 0.1739]
    """
    order = numpy.asarray(order, dtype=int).flatten()
    lower = numpy.asarray(lower).flatten()
    upper = numpy.asarray(upper).flatten()

    dim = max(lower.size, upper.size, order.size)
    order = numpy.ones(dim, dtype=int)*order
    lower = numpy.ones(dim)*lower
    upper = numpy.ones(dim)*upper

    if composite is None:
        composite = numpy.array(0)
    composite = numpy.asarray(composite)

    if not composite.size:
        composite = numpy.array([numpy.linspace(0, 1, composite+1)]*dim)

    else:
        composite = numpy.array(composite)
        if len(composite.shape) <= 1:
            composite = numpy.transpose([composite])
        composite = ((composite.T-lower)/(upper-lower)).T

    results = [_gauss_legendre(order[i], composite[i]) for i in range(dim)]
    abscis = numpy.array([_[0] for _ in results])
    weights = numpy.array([_[1] for _ in results])

    abscis = chaospy.quad.combine(abscis)
    weights = chaospy.quad.combine(weights)

    abscis = (upper-lower)*abscis + lower
    weights = numpy.prod(weights*(upper-lower), 1)

    return abscis.T, weights


def _gauss_legendre(order, composite=1):
    """Backend function."""
    inner = numpy.ones(order+1)*0.5
    outer = numpy.arange(order+1)**2
    outer = outer/(16*outer-4.)

    banded = numpy.diag(numpy.sqrt(outer[1:]), k=-1) + numpy.diag(inner) + \
            numpy.diag(numpy.sqrt(outer[1:]), k=1)
    vals, vecs = numpy.linalg.eig(banded)

    abscis, weight = vals.real, vecs[0, :]**2
    indices = numpy.argsort(abscis)
    abscis, weight = abscis[indices], weight[indices]

    n_abscis = len(abscis)

    composite = numpy.array(composite).flatten()
    composite = list(set(composite))
    composite = [comp for comp in composite if (comp < 1) and (comp > 0)]
    composite.sort()
    composite = [0]+composite+[1]

    abscissas = numpy.empty(n_abscis*(len(composite)-1))
    weights = numpy.empty(n_abscis*(len(composite)-1))
    for dim in range(len(composite)-1):
        abscissas[dim*n_abscis:(dim+1)*n_abscis] = \
            abscis*(composite[dim+1]-composite[dim]) + composite[dim]
        weights[dim*n_abscis:(dim+1)*n_abscis] = \
            weight*(composite[dim+1]-composite[dim])

    return abscissas, weights



"""
Fejér quadrature method.

The same method as :ref:`clenshaw_curtis`, but without the end-points. This
makes this a better method for performing quadrature on infinite intervals, as
the evaluation does not contain illegal values.

Example usage
-------------

The first few orders with linear growth rule::

    >>> distribution = chaospy.Uniform(0, 1)
    >>> for order in [0, 1, 2, 3]:
    ...     X, W = chaospy.generate_quadrature(order, distribution, rule="F")
    ...     print(order, numpy.around(X, 3), numpy.around(W, 3))
    0 [[0.5]] [1.]
    1 [[0.25 0.75]] [0.5 0.5]
    2 [[0.146 0.5   0.854]] [0.286 0.429 0.286]
    3 [[0.095 0.345 0.655 0.905]] [0.188 0.312 0.312 0.188]

The first few orders with exponential growth rule::

    >>> for order in [0, 1, 2]:
    ...     X, W = chaospy.generate_quadrature(
    ...         order, distribution, rule="F", growth=True)
    ...     print(order, numpy.around(X, 2), numpy.around(W, 2))
    0 [[0.5]] [1.]
    1 [[0.15 0.5  0.85]] [0.29 0.43 0.29]
    2 [[0.04 0.15 0.31 0.5  0.69 0.85 0.96]] [0.07 0.14 0.18 0.2  0.18 0.14 0.07]

Applying the rule using Smolyak sparse grid::

    >>> distribution = chaospy.Iid(chaospy.Uniform(0, 1), 2)
    >>> X, W = chaospy.generate_quadrature(
    ...     2, distribution, rule="F", growth=True, sparse=True)
    >>> print(numpy.around(X, 3))
    [[0.5   0.146 0.5   0.854 0.5   0.038 0.146 0.309 0.5   0.691 0.854 0.962
      0.5   0.146 0.5   0.854 0.5  ]
     [0.038 0.146 0.146 0.146 0.309 0.5   0.5   0.5   0.5   0.5   0.5   0.5
      0.691 0.854 0.854 0.854 0.962]]
    >>> print(numpy.around(W, 3))
    [ 0.073  0.071 -0.02   0.071  0.181  0.073 -0.02   0.181 -0.246  0.181
     -0.02   0.073  0.181  0.071 -0.02   0.071  0.073]
"""
from __future__ import division

import numpy
import chaospy.quad


def quad_fejer(order, lower=0, upper=1, growth=False, part=None):
    """
    Generate the quadrature abscisas and weights in Fejér quadrature.

    Example:
        >>> abscissas, weights = quad_fejer(3, 0, 1)
        >>> print(numpy.around(abscissas, 4))
        [[0.0955 0.3455 0.6545 0.9045]]
        >>> print(numpy.around(weights, 4))
        [0.1804 0.2996 0.2996 0.1804]
    """
    order = numpy.asarray(order, dtype=int).flatten()
    lower = numpy.asarray(lower).flatten()
    upper = numpy.asarray(upper).flatten()

    dim = max(lower.size, upper.size, order.size)

    order = numpy.ones(dim, dtype=int)*order
    lower = numpy.ones(dim)*lower
    upper = numpy.ones(dim)*upper

    composite = numpy.array([numpy.arange(2)]*dim)

    if growth:
        results = [
            _fejer(numpy.where(order[i] == 0, 0, 2.**(order[i]+1)-2))
            for i in range(dim)
        ]
    else:
        results = [
            _fejer(order[i], composite[i]) for i in range(dim)
        ]

    abscis = [_[0] for _ in results]
    weight = [_[1] for _ in results]

    abscis = chaospy.quad.combine(abscis, part=part).T
    weight = chaospy.quad.combine(weight, part=part)

    abscis = ((upper-lower)*abscis.T + lower).T
    weight = numpy.prod(weight*(upper-lower), -1)

    assert len(abscis) == dim
    assert len(weight) == len(abscis.T)

    return abscis, weight


def _fejer(order, composite=None):
    r"""
    Backend method.

    Examples:
        >>> print(*_fejer(0))
        [0.5] [1.]
        >>> print(*_fejer(1))
        [0.25 0.75] [0.44444444 0.44444444]
        >>> print(*_fejer(2))
        [0.14644661 0.5        0.85355339] [0.26666667 0.4        0.26666667]
        >>> print(*_fejer(3), sep="\n")
        [0.0954915 0.3454915 0.6545085 0.9045085]
        [0.18037152 0.29962848 0.29962848 0.18037152]
        >>> print(*_fejer(4), sep="\n")
        [0.0669873 0.25      0.5       0.75      0.9330127]
        [0.12698413 0.22857143 0.26031746 0.22857143 0.12698413]
        >>> print(*_fejer(5), sep="\n")
        [0.04951557 0.1882551  0.38873953 0.61126047 0.8117449  0.95048443]
        [0.0950705  0.17612121 0.2186042  0.2186042  0.17612121 0.0950705 ]
    """
    order = int(order)
    if order == 0:
        return numpy.array([.5]), numpy.array([1.])

    order += 2

    theta = (order-numpy.arange(order+1))*numpy.pi/order
    abscisas = 0.5*numpy.cos(theta) + 0.5

    N, K = numpy.mgrid[:order+1, :order//2]
    weights = 2*numpy.cos(2*(K+1)*theta[N])/(4*K*(K+2)+3)
    if order % 2 == 0:
        weights[:, -1] *= 0.5
    weights = (1-numpy.sum(weights, -1)) / order

    return abscisas[1:-1], weights[1:-1]



"""Frontend for the generation of quadrature rules."""
import inspect

import numpy as np
from scipy.misc import comb

from . import collection, sparse_grid


def generate_quadrature(order, domain, accuracy=100, sparse=False, rule="C",
                        composite=1, growth=None, part=None, **kws):
    """
    Numerical quadrature node and weight generator.

    Args:
        order (int):
            The order of the quadrature.
        domain (numpy.ndarray, Dist):
            If array is provided domain is the lower and upper bounds (lo,up).
            Invalid if gaussian is set.  If Dist is provided, bounds and nodes
            are adapted to the distribution. This includes weighting the nodes
            in Clenshaw-Curtis quadrature.
        accuracy (int):
            If gaussian is set, but the Dist provieded in domain does not
            provide an analytical TTR, ac sets the approximation order for the
            descitized Stieltje's method.
        sparse (bool):
            If True used Smolyak's sparse grid instead of normal tensor product
            grid.
        rule (str):
            Rule for generating abscissas and weights. Either done with
            quadrature rules, or with random samples with constant weights.
        composite (int):
            If provided, composite quadrature will be used.  Value determines
            the number of domains along an axis. Ignored in the case
            gaussian=True.
        growth (bool):
            If True sets the growth rule for the composite quadrature rule to
            exponential for Clenshaw-Curtis quadrature.
    """
    from ..distributions.baseclass import Dist
    isdist = isinstance(domain, Dist)
    if isdist:
        dim = len(domain)
    else:
        dim = np.array(domain[0]).size

    rule = rule.lower()
    if len(rule) == 1:
        rule = collection.QUAD_SHORT_NAMES[rule]

    quad_function = collection.get_function(
        rule,
        domain,
        growth=growth,
        composite=composite,
        accuracy=accuracy,
    )

    if sparse:
        order = np.ones(len(domain), dtype=int)*order
        abscissas, weights = sparse_grid.sparse_grid(quad_function, order, dim)

    else:
        abscissas, weights = quad_function(order)

    assert len(weights) == abscissas.shape[1]
    assert len(abscissas.shape) == 2

    if isdist:
        if rule != "golub_welsch":
            if dim == 1:
                weights *= domain.pdf(abscissas).flatten()
            else:
                weights *= domain.pdf(abscissas)
        if not sparse:
            weights /= np.sum(weights)

    return abscissas, weights



"""Wrapper to the OpenTURNS distribution."""
import numpy

from ..baseclass import Dist


class OTDistribution(Dist):
    """OpenTURNS distribution."""

    def __init__(self, distribution):
        """
        Args:
            distribution (openturns.Distribution):
                1D distribution created in OpenTURNS.
        """
        Dist.__init__(self)
        if distribution.getDimension() != 1:
            raise Exception("Only 1D OpenTURNS distribution are supported for now")
        self.distribution = distribution

    def _pdf(self, x):
        return numpy.array(self.distribution.computePDF(numpy.atleast_2d(x).T).asPoint())

    def _cdf(self, x):
        return numpy.array(self.distribution.computeCDF(numpy.atleast_2d(x).T).asPoint())

    def _ppf(self, q):
        return numpy.array(self.distribution.computeQuantile(q[0]).asPoint())

    def _bnd(self):
        rng = self.distribution.getRange()
        return rng.getLowerBound()[0], rng.getUpperBound()[0]

    def _mom(self, k):
        return self.getMoment(k)[0]

    def _str(self):
        return self.distribution.__str__()



"""Arc-Sinus Hyperbolic."""
import numpy
from ..baseclass import Dist
from .. import evaluation, approximation


class Arcsinh(Dist):
    """
    Arc-Sinus Hyperbolic.

    Args:
        dist (Dist): Distribution to perform transformation on.

    Example:
        >>> distribution = chaospy.Arcsinh(chaospy.Uniform(0, 1))
        >>> print(distribution)
        Arcsinh(Uniform(lower=0, upper=1))
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [0.1987 0.39   0.5688 0.7327]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [1.5907 1.7404 2.0115 2.4324]
        >>> print(numpy.around(distribution.sample(4), 4))
        [0.6142 0.1148 0.8458 0.4652]
        >>> print(numpy.around(distribution.mom(1), 4))
        0.4917
        >>> print(numpy.around(distribution.ttr([0, 1, 2]), 4))
        [[0.8479 0.2254 0.5361]
         [1.     0.3332 0.7026]]
    """

    def __init__(self, dist):
        assert isinstance(dist, Dist)
        Dist.__init__(self, dist=dist)

    def _pdf(self, x, dist, cache):
        """Probability density function."""
        return evaluation.evaluate_density(
            dist, numpy.sinh(x), cache=cache)*numpy.cosh(1+x*x)

    def _cdf(self, x, dist, cache):
        return evaluation.evaluate_forward(dist, numpy.sinh(x), cache=cache)

    def _ppf(self, q, dist, cache):
        return numpy.arcsinh(evaluation.evaluate_inverse(dist, q, cache=cache))

    def _bnd(self, x, dist, cache):
        return numpy.arcsinh(evaluation.evaluate_bound(
            dist, numpy.sinh(x), cache=cache))

    def _mom(self, x, dist, cache):
        return approximation.approximate_moment(self, x, cache=cache)

    def __len__(self):
        return len(self.prm["dist"])

    def __str__(self):
        return self.__class__.__name__ + "(" + str(self.prm["dist"]) + ")"

    def _fwd_cache(self, cache):
        dist = evaluation.get_forward_cache(self.prm["dist"], cache)
        if not isinstance(dist, Dist):
            return numpy.arcsinh(dist)
        return self

    def _inv_cache(self, cache):
        dist = evaluation.get_forward_cache(self.prm["dist"], cache)
        if not isinstance(dist, Dist):
            return numpy.sinh(dist)
        return self



import numpy as np
from _cubature import _cubature

def cubature(ndim, func, xmin, xmax, args=None, adaptive='h',
             abserr=1.e-8, relerr=1.e-8, norm=0, maxEval=0,
             vectorized=True):
    """Numerical-integration using cubature technique.

    Parameters
    ----------

    ndim : integer
        Number dimensions or number of variables being integrated.
    func : function
        If ``vectorized=False`` the function must have the form:
            ``f(x_array, *args)``
            - `x_array` in an array containing the `ndim` variables
               being integrated
            - `args` is a tuple containing any other arguments
              required by the function
            - the function must return a 1-D `np.ndarray` object
            - example:
                ``def func(x_array, *args):``
                ``    x, y = x_array``
                ``    return np.array([x**2-y**2, x*y, x*y**2])``
        and must return a 1-D ``np.ndarray`` object.
        If ``vectorized=True`` the function must have the form:
            ``f(x_array, npt, *args)``
            - `x_array` has ``shape[0]=ndim*npt``
            - `npt` tells the number of points passed to the function
            - `args` is a tuple containing any other arguments
              required by the function
            - example:
                ``def func(x_array, npt, *args):``
                ``    ndim=2``
                ``    fdim=3``
                ``    out = np.zeros(fdim*npt)``
                ``    for i in range(npt):
                ``        x = x_array[i*ndim]``
                ``        y = x_array[i*ndim+1]``
                ``        out[i*fdim] = x**2-y**2``
                ``        out[i*fdim+1] = x*y``
                ``        out[i*fdim+2] = x*y**2``
                ``    return out``
        The results from both vectorized and non-vetorized examples
        above will be the same, but the vectorized implementation
        allows more control for the user to parallelize the computation
        of the integration points.
    xmin : numpy.ndarray
        A 1-D array carring the minimum integration limit for each
        variable being integrated. It must be have:
        ``xmin.shape[0]=ndim``.
    xmax : numpy.ndarray
        A 1-D array carring the maximum integration limit for each
        variable being integrated. It must be have:
        ``xmax.shape[0]=ndim``.
    args : tuple or list
        Contains the extra arguments required by `func`
        (default = None)
    adaptive : string
        The adaptive scheme used along the adaptive integration.
        'h' means 'h-adaptive', where the domain is partitioned
        'p' means 'p-adaptive', where the order of the integration rule
                                is increased
        The 'p-adaptive' scheme is often better for smoth functions in
        low dimensions.
        (default = 'h')
    epsabs : double
        The maximum number of function evaluations.
        (default = 1.49e-08)
    epsrel : double
        The maxmum number of function evaluations.
        (default = 1.49e-08)
    norm : integer
        Specifies the norm that is used to measure the error and
        determine convergence properties (irrelevant for
        single-valued functions).
        The `norm` argument takes one of the values:
        - 0 (ERROR_INDIVIDUAL) convergence is achieved only when each
            integrand individually satisfies the requested error
            tolerances;
        - 1 (ERROR_PAIRED) like ERROR_INDIVIDUAL, except that the
            integrands are grouped into consecutive pairs, with the error
            tolerance applied in a L2 sense to each pair. This option is
            mainly useful for integrating vectors of complex numbers,
            where each consecutive pair of real integrans is the real
            and imaginary parts of a single complex integrand, and you
            only care about the error in the complex plane rather than
            the error in the real and imaginary parts separately;
        - 2 (ERROR_L2)
        - 3 (ERROR_L1)
        - 4 (ERROR_LINF)
            the absolute error is measured as |e| and the relative error
            as |err|/|val|, where |...| is the L1, L2, or L-infinity
            norm, respectively.  (|x| in the L1 norm is the sum of the
            absolute values of the components, in the L2 norm is the
            root mean square of the components, and in the L-infinity
            norm is the maximum absolute value of the components).
        (default = 0)
    maxEval : integer
        The maximum number of function evaluations.
        (default = 0, which means unlimited)
    vectorized : boolean
        If ``vectorized=True`` the integration points are passed to the
        integrand function as an array of points, allowing parallel
        evaluation of different points.
        (default = False)

    Returns
    -------

    val : numpy.ndarray
        The 1-D array of length ``fdim`` with the computed integral values

    err : numpy.ndarray
        The 1-D array of length ``fdim`` with the estimated errors. For
        smooth functions this estimate is usually conservative (see the
        results from the ``test_cubature.py`` script.

    Notes
    -----
    * The supplied function must return a 1-D ``np.ndarray`` object,
      even for single-valued functions

    References
    ----------
    .. [1] `Cubature (Multi-dimensional integration)
           <http://ab-initio.mit.edu/wiki/index.php/Cubature>`_.

    Examples
    --------
    >>> import numpy as np
    >>> from cubature import cubature

    Volume of a sphere

    >>>
    >>> ndim = 3
    >>> xmin = np.zeros(ndim)
    >>> def integrand_sphere(x_array, *args):
    ...     r, theta, phi = x_array
    ...     return np.array([r**2*sin(phi)])
    >>>
    >>> radius = 1.
    >>> xmin = np.array([0, 0, 0], float)
    >>> xmax = np.array([radius, 2*pi, pi], float)
    >>> val, err = cubature(3, integrand_sphere, xmin, xmax)

    More examples in ./examples/*.py

    """
    if not args:
        args = ()
    # checking xmin and xmax
    try:
        xmin = np.array(xmin, np.float64)
        xmin = (xmin.T*np.ones(ndim)).T
    except:
        raise ValueError('xmin.shape does not fit ndim')
    try:
        xmax = np.array(xmax, np.float64)
        xmax = (xmax.T*np.ones(ndim)).T
        assert xmax.shape[0] == ndim
    except:
        raise ValueError('xmax.shape does not fit ndim')

    # getting fdim and checking function
    dummy_x = np.zeros((ndim), dtype=np.float64)
    try:
        if vectorized:
            val = func(dummy_x, 1, *args)
        else:
            val = func(dummy_x, *args)
    except:
        # this print was kept to give the user some more info
        print(func(dummy_x, *args))
        raise ValueError('Invalid input function!')

    val = np.array(val)
    try:
        assert val.ndim == 1
    except:
        raise ValueError('The input function must return a 1-D array!')
    fdim = val.shape[0]
    # initializing val and err
    val = np.zeros((fdim), dtype=np.float64)
    err = np.zeros((fdim), dtype=np.float64)
    # calling the integrator
    ans = _cubature(func, fdim, xmin, xmax, args, adaptive,
                    abserr, relerr, norm, maxEval,
                    vectorized, val, err)
    return val, err

#TODO
# - implement multiprocessing dividing the integration interval and spawning
# - a thread for each sub-interval...
# - perform a cProfile to see where the bottle nech actually is

if __name__=="__main__":
    import doctest
    from numpy import pi, sin
    doctest.testmod()




"""
As the number of dimensions increases linear, the number of samples increases
exponentially. This is known as the curse of dimensionality. Except for
switching to Monte Carlo integration, the is no way to completly guard against
this problem. However, there are some possibility to mitigate the problem
personally. One such strategy is to employ Smolyak sparse-grid quadrature. This
method uses a quadrature rule over a combination of different orders to tailor
a scheme that uses fewer abscissas points than a full tensor-product approach.

To use Smolyak sparse-grid in ``chaospy``, just pass the flag ``sparse=True``
to the ``generate_quadrature`` function. For example::

    >>> distribution = chaospy.J(chaospy.Uniform(0, 4), chaospy.Uniform(0, 4))
    >>> X, W = chaospy.generate_quadrature(3, distribution, sparse=True)
    >>> print(numpy.around(X, 4))
    [[0. 2. 4. 2. 0. 1. 2. 3. 4. 2. 0. 2. 4.]
     [0. 0. 0. 1. 2. 2. 2. 2. 2. 3. 4. 4. 4.]]
    >>> print(numpy.around(W, 4))
    [-0.0833  0.2222 -0.0833  0.4444  0.2222  0.4444 -0.6667  0.4444  0.2222
      0.4444 -0.0833  0.2222 -0.0833]

This compared to the full tensor-product grid::

    >>> X, W = chaospy.generate_quadrature(3, distribution)
    >>> print(numpy.around(X, 4))
    [[0. 0. 0. 0. 1. 1. 1. 1. 3. 3. 3. 3. 4. 4. 4. 4.]
     [0. 1. 3. 4. 0. 1. 3. 4. 0. 1. 3. 4. 0. 1. 3. 4.]]
    >>> print(numpy.around(W, 4))
    [0.0031 0.0247 0.0247 0.0031 0.0247 0.1975 0.1975 0.0247 0.0247 0.1975
     0.1975 0.0247 0.0031 0.0247 0.0247 0.0031]

The method works with all quadrature rules, but is known to be quite
inefficient when applied to rules that can not be nested. For example using
Gauss-Legendre samples::

    >>> X, W = chaospy.generate_quadrature(
    ...     3, distribution, rule="E", sparse=True)
    >>> print(len(W))
    119
    >>> X, W = chaospy.generate_quadrature(
    ...     3, distribution, rule="E", sparse=False)
    >>> print(len(W))
    64
"""

import numpy
from scipy.special import comb

import chaospy.bertran


def sparse_grid(func, order, dim=None, skew=None):
    """
    Smolyak sparse grid constructor.

    Args:
        func (:py:data:typing.Callable):
            Function that takes a single argument ``order`` of type
            ``numpy.ndarray`` and with ``order.shape = (dim,)``
        order (int, numpy.ndarray):
            The order of the grid. If ``numpy.ndarray``, it overrides both
            ``dim`` and ``skew``.
        dim (int):
            Number of dimension.
        skew (list):
            Order skewness.
    """
    if not isinstance(order, int):
        orders = numpy.array(order).flatten()
        dim = orders.size
        m_order = int(numpy.min(orders))
        skew = [order-m_order for order in orders]
        return sparse_grid(func, m_order, dim, skew)

    abscissas, weights = [], []
    bindex = chaospy.bertran.bindex(order-dim+1, order, dim)

    if skew is None:
        skew = numpy.zeros(dim, dtype=int)
    else:
        skew = numpy.array(skew, dtype=int)
        assert len(skew) == dim

    for idx in range(
            chaospy.bertran.terms(order, dim)
            - chaospy.bertran.terms(order-dim, dim)):

        idb = bindex[idx]
        abscissa, weight = func(skew+idb)
        weight *= (-1)**(order-sum(idb))*comb(dim-1, order-sum(idb))
        abscissas.append(abscissa)
        weights.append(weight)

    abscissas = numpy.concatenate(abscissas, 1)
    weights = numpy.concatenate(weights, 0)

    abscissas = numpy.around(abscissas, 15)
    order = numpy.lexsort(tuple(abscissas))
    abscissas = abscissas.T[order].T
    weights = weights[order]

    # identify non-unique terms
    diff = numpy.diff(abscissas.T, axis=0)
    unique = numpy.ones(len(abscissas.T), bool)
    unique[1:] = (diff != 0).any(axis=1)

    # merge duplicate nodes
    length = len(weights)
    idx = 1
    while idx < length:
        while idx < length and unique[idx]:
            idx += 1
        idy = idx+1
        while idy < length and not unique[idy]:
            idy += 1
        if idy-idx > 1:
            weights[idx-1] = numpy.sum(weights[idx-1:idy])
        idx = idy+1

    abscissas = abscissas[:, unique]
    weights = weights[unique]

    return abscissas, weights



"""Gaussian kernel density estimation."""
import numpy
from scipy.special import ndtr, ndtri
from scipy.stats import gaussian_kde

from ..baseclass import Dist


class GaussianKDE(Dist):
    """
    Gaussian kernel density estimation.

    Examples:
        >>> distribution = GaussianKDE([0, 1, 1, 1, 2])
        >>> print(distribution)
        GaussianKDE()
        >>> q = numpy.linspace(0, 1, 7)[1:-1]
        >>> print(numpy.around(q, 4))
        [0.1667 0.3333 0.5    0.6667 0.8333]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        >>> print(numpy.around(distribution.inv(q), 4))
        >>> print(numpy.around(distribution.sample(4), 4))
        >>> print(numpy.around(distribution.mom(1), 4))
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
    """

    def __init__(self, samples, lower=None, upper=None):

        samples = numpy.asarray(samples)
        if len(samples.shape) == 1:
            samples = samples.reshape(1, -1)
        kernel = gaussian_kde(samples, bw_method="scott")

        if lower is None:
            lower = samples.min(axis=-1)
        if upper is None:
            upper = samples.max(axis=-1)

        self.lower = lower
        self.upper = upper
        self.samples = samples
        self.l_fwd = numpy.linalg.cholesky(kernel.covariance)
        self.l_inv = numpy.linalg.inv(self.l_fwd)
        super(GaussianKDE, self).__init__()

    def __len__(self):
        return len(self.samples)

    def _pdf(self, x_data):
        out = numpy.zeros(x_data.shape)

        # first dimension is simple:
        for sample in self.samples[0]:
            z = self.l_inv[0, 0]*(x_data[0] - sample)
            out[0] += numpy.e**(-0.5*z*z) / len(self.samples[0])
        out[0] *= self.l_inv[0, 0]*(2*numpy.pi)**-0.5

        return out

    def _cdf(self, x_data):
        out = numpy.zeros(x_data.shape)

        # first dimension is simple:
        print(x_data)
        for sample in self.samples[0]:
            z = self.l_inv[0, 0]*(x_data[0] - sample)
            print(z)
            z = ndtr(z)
            out[0] += z / len(self.samples[0])

        return out

    def _ppf(self, u_data):
        out = numpy.zeros(u_data.shape)

        # first dimension is simple:
        z = self.l_fwd[0, 0]*ndtri(u_data[0])
        print(u_data)
        for sample in self.samples[0]:
            print(z + sample)
            out[0] += (z + sample) / len(self.samples[0])
            # print("i", out)
        # print(out)

        return out

    def _bnd(self, x_data):
        return self.lower, self.upper



"""
Laja quadrature is a newer method for performing quadrature in stochastical
problems. The method is described in a `journal paper`_ by Narayan and Jakeman.

.. _journal paper: https://arxiv.org/pdf/1404.5663.pdf

Example usage
-------------

The first few orders::

    >>> distribution = chaospy.Uniform(0, 1)
    >>> for order in [0, 1, 2, 3, 4]:
    ...     abscissas, weights = chaospy.generate_quadrature(
    ...         order, distribution, rule="J")
    ...     print(order, numpy.around(abscissas, 3), numpy.around(weights, 3))
    0 [[0.5]] [1.]
    1 [[0.5 1. ]] [1. 0.]
    2 [[0.  0.5 1. ]] [0.167 0.667 0.167]
    3 [[0.    0.5   0.789 1.   ]] [0.167 0.667 0.    0.167]
    4 [[0.    0.171 0.5   0.789 1.   ]] [0.043 0.289 0.316 0.28  0.072]
"""
import numpy
from scipy.optimize import fminbound

import chaospy.quad


def quad_leja(order, dist):
    """
    Generate Leja quadrature node.

    Example:
        >>> abscisas, weights = quad_leja(3, chaospy.Normal(0, 1))
        >>> print(numpy.around(abscisas, 4))
        [[-2.7173 -1.4142  0.      1.7635]]
        >>> print(numpy.around(weights, 4))
        [0.022  0.1629 0.6506 0.1645]
    """
    from chaospy.distributions import evaluation
    if len(dist) > 1 and evaluation.get_dependencies(*list(dist)):
        raise evaluation.DependencyError(
            "Leja quadrature do not supper distribution with dependencies.")

    if len(dist) > 1:
        if isinstance(order, int):
            out = [quad_leja(order, _) for _ in dist]
        else:
            out = [quad_leja(order[_], dist[_]) for _ in range(len(dist))]

        abscissas = [_[0][0] for _ in out]
        weights = [_[1] for _ in out]
        abscissas = chaospy.quad.combine(abscissas).T
        weights = chaospy.quad.combine(weights)
        weights = numpy.prod(weights, -1)

        return abscissas, weights

    lower, upper = dist.range()
    abscissas = [lower, dist.mom(1), upper]
    for _ in range(int(order)):

        obj = create_objective(dist, abscissas)
        opts, vals = zip(
            *[fminbound(
                obj, abscissas[idx], abscissas[idx+1], full_output=1)[:2]
              for idx in range(len(abscissas)-1)]
        )
        index = numpy.argmin(vals)
        abscissas.insert(index+1, opts[index])

    abscissas = numpy.asfarray(abscissas).flatten()[1:-1]
    weights = create_weights(abscissas, dist)
    abscissas = abscissas.reshape(1, abscissas.size)

    return numpy.array(abscissas), numpy.array(weights)


def create_objective(dist, abscissas):
    """Create objective function."""
    abscissas_ = numpy.array(abscissas[1:-1])
    def obj(absisa):
        """Local objective function."""
        out = -numpy.sqrt(dist.pdf(absisa))
        out *= numpy.prod(numpy.abs(abscissas_ - absisa))
        return out
    return obj


def create_weights(nodes, dist):
    """Create weights for the Laja method."""
    poly = chaospy.quad.generate_stieltjes(dist, len(nodes)-1, retall=True)[0]
    poly = chaospy.poly.flatten(chaospy.poly.Poly(poly))
    weights_inverse = poly(nodes)
    weights = numpy.linalg.inv(weights_inverse)
    return weights[:, 0]



import sys
import os
from subprocess import Popen, PIPE
from operator import mod

import numpy as np
from numpy import pi, sin, cos, exp, zeros, ones, float64

from cubature import cubature

count = 0
fdim = 1
function = 0
ndim = 0
K_2_SQRTPI = 1.12837916709551257390
radius = 0.50124145262344534123412

class Logger(object):
    def __init__(self):
        self.file = open('test_cubature.txt', 'w')
        self.file.close()
        self.stdout = sys.stdout
        sys.stdout = self
    def write(self, text):
        self.file = open('test_cubature.txt', 'a')
        self.file.write(text)
        self.file.close()
        self.stdout.write(text)

def f_test(x, *args):
    global count, ndim, fdim, function, K_2_SQRTPI, radius
    count += 1
    if function == 0:
        val = ones(fdim, float64)
        for i in range(ndim):
            val *= cos(x[i])
    elif function == 1:
        scale = 1.
        val = zeros(fdim, float64)
        for i in range(ndim):
            if x[i] > 0:
                z = (1 - x[i]) / x[i]
                val += z**2
                scale *= K_2_SQRTPI / x[i]**2
            else:
                scale = 0
        val = scale * exp(-val)
    elif function == 2:
        val = zeros(fdim, float64)
        for i in range(ndim):
            val += x[i]**2
        val[:] = (val[0] < radius**2)
    elif function == 3:
        val = ones(fdim, float64)
        for i in range(ndim):
            val *= 2.0 * x[i]
    elif function == 4:
        sum1 = zeros(fdim, float64)
        sum2 = zeros(fdim, float64)
        a = 0.1
        for i in range(ndim):
            dx1 = x[i] - 1. / 3.
            dx2 = x[i] - 2. / 3.
            sum1 += dx1**2
            sum2 += dx2**2
        return 0.5 * pow (K_2_SQRTPI / (2. * a), ndim) \
                   * (exp (-sum1 / a**2) + exp (-sum2 / a**2))
    elif function == 5:
        sum1 = zeros(fdim, float64)
        sum2 = zeros(fdim, float64)
        a = 0.1
        for i in range(ndim):
            dx1 = x[i] - 1. / 3.
            dx2 = x[i] - 2. / 3.
            sum1 += dx1 * dx1
            sum2 += dx2 * dx2
        return 0.5 * pow (K_2_SQRTPI / (2. * a), ndim) \
                   * (exp (-sum1 / a**2) + exp (-sum2 / a**2))
    elif function == 6:
        val = ones(fdim, float64)
        c = (1.+ 10.**0.5)/9.
        for i in range(ndim):
            val *= c / (c + 1) * pow((c + 1) / (c + x[i]), 2.0)
    elif function == 7:
        p = ones(fdim, float64) / ndim
        val = pow(1 + p, ndim)
        for i in range(ndim):
            val *= pow(x[i], p)
    return val

def f_test_vec(xvec, npt, *args):
    global ndim, fdim
    out = np.zeros(fdim*npt)
    x = np.zeros(ndim)
    for i in range(npt):
        for j in range(ndim):
            x[j] = xvec[i*ndim+j]
        out[i*fdim:(i+1)*fdim] = f_test(x, *args)
    return out


def exact0(ndim, xmax):
    val = 1.
    for i in range(ndim):
        val *= sin(xmax[i])
    return val

def exact2_S(n):
     fact = 1
     if mod(n, 2) == 0:
         val = 2 * pow(pi, n * 0.5)
         n = n / 2
         while (n > 1):
             fact *= n
             n -= 1
         val /= fact
     else:
         val = (1 << (n/2 + 1)) * pow(pi, n/2)
         while (n > 2):
             fact *= n
             n -= 2
         val /= fact
     return val

def exact2(ndim, xmax):
    global radius
    val = 1 if ndim == 0 else exact2_S(ndim) * pow(radius * 0.5, ndim) / ndim
    return val

def exact_integral(i, ndim, xmax):
    if i in [1,3,4,5,6,7]:
        return 1.
    elif i == 0:
        return exact0(ndim, xmax)
    elif i == 2:
        return exact2(ndim, xmax)

def main(lndim, tol, functions, maxEval, lfdim):
    global count, ndim, fdim, function
    ndim = lndim
    fdim = lfdim
    logger = Logger() # instanciate Logger to redirect print to test_cubature.txt
    xmin = zeros(ndim)
    xmax = ones(ndim)
    abserr = tol
    relerr = tol

    for vectorized in [False, True]:
        print('======================================')
        print('           VECTORIZED={0}'.format(vectorized))
        print('======================================')
        print('')
        for function in functions:
            count = 0
            print('______________________________________')
            print('                                      ')
            print('                CASE {0}'.format(function))
            print('______________________________________')
            for adaptive in ['h', 'p']:
                if adaptive == 'h':
                    print('__________________')
                    print('                  ')
                    print('Testing h_adaptive')
                    print('__________________')
                else:
                    print('__________________')
                    print('                  ')
                    print('Testing p_adaptive')
                    print('__________________')
                print('')
                print('Python Cubature:')
                print('----------------')
                print('{0}-dim integral, tolerance = {1}'.format(ndim,
                                                                 relerr))
                print('')
                if vectorized:
                    val, err = cubature(ndim, f_test_vec, xmin, xmax, (),
                             adaptive, abserr, relerr, norm=0,
                             maxEval=maxEval, vectorized=vectorized)
                else:
                    val, err = cubature(ndim, f_test, xmin, xmax, (),
                             adaptive, abserr, relerr, norm=0,
                             maxEval=maxEval, vectorized=vectorized)
                true_err = abs(val[0] - exact_integral(function, ndim, xmax))
                print('integrand {0}: integral = {1}, est err = {2}, true err = {3:e}'
                      .format(function, val[0], err[0], true_err))
                print('')
                print('#evals = {0}'.format(count))
                print('')

                htest_path = os.path.join('.', 'cpackage', 'htest.exe')
                ptest_path = os.path.join('.', 'cpackage', 'ptest.exe')
                if os.path.isfile(htest_path) or os.path.isfile(ptest_path):
                    print('C Cubature:')
                    print('-----------')
                else:
                    print('C Cubature program not compiled!')
                    print('compile using (more details in ".\cpackage\README:"')
                    print('\tcc -DHCUBATURE -o htest hcubature.c test.c -lm')
                    print('\tcc -DPCUBATURE -o ptest pcubature.c test.c -lm')
                fdim_str = '/'.join(['x' for j in range(fdim)])
                if adaptive=='h' and os.path.isfile(htest_path):
                    p = Popen([htest_path] +
                          list(map(str, [ndim,tol,function,
                                         maxEval,fdim_str])),
                          stdout = PIPE)
                    p.wait()
                    for l in p.stdout: print(l)
                if adaptive=='p' and os.path.isfile(ptest_path):
                    p = Popen([ptest_path] +
                          list(map(str, [ndim,tol,function,
                                         maxEval,fdim_str])),
                          stdout = PIPE)
                    p.wait()
                    for l in p.stdout: print(l)

if __name__ == '__main__':
    from __init__ import run_test
    run_test()



"""
Uncertainty Quantification Toolbox
==================================

This module contains tools for performing uncertainty quantification of models.
"""
import logging
import os

from chaospy.version import __version__

import chaospy.bertran
import chaospy.chol
import chaospy.descriptives
import chaospy.distributions
import chaospy.orthogonal
import chaospy.poly
import chaospy.spectral
import chaospy.quad
import chaospy.saltelli
import chaospy.regression

from chaospy.distributions import *
from chaospy.orthogonal import *
from chaospy.poly import *
from chaospy.spectral import *
from chaospy.quad import *
from chaospy.saltelli import *
from chaospy.descriptives import *
from chaospy.regression import *

LOGPATH = os.environ.get("CHAOSPY_LOGPATH", os.devnull)
logging.basicConfig(level=logging.DEBUG, filename=LOGPATH, filemode="w")
streamer = logging.StreamHandler()
streamer.setLevel(logging.WARNING)
logger = logging.getLogger(__name__)
logger.addHandler(streamer)
